// Die Bibliotheken Sanford.Multimedia und Sanford.Multimedia.Midi neben die cs-Dateien legen,
// siehe http://www.lesliesanford.com/Programming/Assemblies.shtml
// Ebenso PortAudio und PortAudioSharp.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using PortAudioSharp;
using Sanford.Multimedia;
using Sanford.Multimedia.Midi;

namespace PortAudioDemo
{
    public partial class Form1 : Form
    {
        private InputDevice inDevice;
        private IntPtr stream;
        private PortAudio.PaStreamCallbackDelegate audioCallback;

        double[] frequencies = new double[128];

        public Form1()
        {
            InitializeComponent();

            for (int i = 0; i < 128; i++)
            {
                frequencies[i] = 440.0 * Math.Pow(2.0, (i - 69) / 12.0);
            }

            audioCallback = new PortAudio.PaStreamCallbackDelegate(work);

            CheckError(PortAudio.Pa_Initialize());

            // Unntigerweise muss man wohl hier auch ein Input angeben.
            PortAudio.PaStreamParameters inputParameters = new PortAudio.PaStreamParameters();
            inputParameters.channelCount = 2;
            inputParameters.device = PortAudio.Pa_GetDefaultInputDevice();
            inputParameters.sampleFormat = PortAudio.PaSampleFormat.paFloat32;

            PortAudio.PaStreamParameters outputParameters = new PortAudio.PaStreamParameters();
            outputParameters.channelCount = 2;
            outputParameters.device = PortAudio.Pa_GetDefaultOutputDevice();
            outputParameters.sampleFormat = PortAudio.PaSampleFormat.paFloat32;

            CheckError(PortAudio.Pa_IsFormatSupported(ref inputParameters, ref outputParameters, 48000.0));
            CheckError(PortAudio.Pa_OpenStream(out stream, ref inputParameters, ref outputParameters, 48000.0,
                2048U, PortAudio.PaStreamFlags.paNoFlag, audioCallback, IntPtr.Zero));
            CheckError(PortAudio.Pa_StartStream(stream));

            inDevice = new InputDevice(0);
            inDevice.ChannelMessageReceived += HandleChannelMessageReceived;
            inDevice.StartRecording();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            inDevice.StopRecording();

            CheckError(PortAudio.Pa_StopStream(stream));
            stream = IntPtr.Zero;
            CheckError(PortAudio.Pa_Terminate());

            base.OnClosing(e);
        }

        private List<KeyPressData> keys = new List<KeyPressData>();

        private void HandleChannelMessageReceived(object sender, ChannelMessageEventArgs e)
        {
            if (e.Message.Command == ChannelCommand.NoteOn)
            {
                if (e.Message.Data2 > 0)
                {
                    KeyPressData k = new KeyPressData();
                    k.note = e.Message.Data1;
                    k.pressed = true;
                    lock (keys)
                    {
                        bool found = false;
                        for (int i = 0; i < keys.Count; i++)
                        {
                            if (keys[i].note == e.Message.Data1)
                            {
                                keys[i].pressed = true;
                                found = true;
                                break;
                            }
                        }
                        if (!found)
                        {
                            keys.Add(k);
                        }
                    }
                }
                else
                {
                    lock (keys)
                    {
                        for (int i = 0; i < keys.Count; i++)
                        {
                            if (keys[i].note == e.Message.Data1)
                            {
                                keys[i].pressed = false;
                                break;
                            }
                        }
                    }
                }
            }
            else if (e.Message.Command == ChannelCommand.NoteOff)
            {
                lock (keys)
                {
                    for (int i = 0; i < keys.Count; i++)
                    {
                        if (keys[i].note == e.Message.Data1)
                        {
                            keys[i].pressed = false;
                            break;
                        }
                    }
                }
            }
        }

        private double t = 0.0;

        private PortAudio.PaStreamCallbackResult work(IntPtr dummy, IntPtr output, uint framesPerBuffer, ref PortAudio.PaStreamCallbackTimeInfo timeInfo, PortAudio.PaStreamCallbackFlags flags, IntPtr userData)
        {
            unsafe
            {
                float* outbuffer = (float*)output;

                for (int s = 0; s < framesPerBuffer; s++)
                {
                    // Vibrato: 10 Hz
                    double vib = Math.Sin(2.0 * Math.PI * 10.0 * t);

                    float sum = 0.0f;
                    lock (keys)
                    {
                        int i = 0;
                        while (i < keys.Count)
                        {
                            if (keys[i].pressed)
                            {
                                keys[i].envelope += 0.5 / 48000.0; // halbe Sekunde bis Ton voll da
                                if (keys[i].envelope >= 1.0)
                                {
                                    keys[i].envelope = 1.0;
                                }
                            }
                            else
                            {
                                keys[i].envelope *= 0.9999; // Ausklingen
                                if (keys[i].envelope < 0.01) // bei -40 dB ausschalten
                                {
                                    keys.RemoveAt(i);
                                    continue;
                                }
                            }

                            double f = 4.0 * frequencies[keys[i].note];
                            double v = keys[i].envelope*Math.Sin(2.0 * Math.PI * f *( t + 0.0003*vib));
                            sum += (float)v;

                            i++;
                        }
                    }

                    // unprofessionell: t wchst (theoretisch) ber alle Grenzen
                    t += 1.0/48000.0;

                    outbuffer[2 * s + 1] = outbuffer[2 * s + 0] = 0.1f*sum;
                }
            }

            return PortAudio.PaStreamCallbackResult.paContinue;
        }

        private static void CheckError(PortAudio.PaErrorCode error)
        {
            if (error != PortAudio.PaErrorCode.paNoError)
            {
                throw new Exception(PortAudio.Pa_GetErrorText(error));
            }
        }
    }

    class KeyPressData
    {
        public int note;
        public bool pressed;
        public double envelope;
    }
}